import logging

from kafka.errors import OffsetOutOfRangeError

from ..config.kafka_config import KafkaConfig

log = logging.getLogger(__name__)


class Messages:
    kafka_obj = KafkaConfig()

    def obtiene_mensajes(self, topico, grupo):
        mensajes = []
        try:
            consumer = self.kafka_obj.get_message(topico, grupo)
            for message in consumer:
                mensajes.append(message.value)
                consumer.commit()

        except OffsetOutOfRangeError:
            # consumer has no idea where they are
            self.kafka_obj.consumer.seek_to_end()
            log.exception("Kafka offset out of range error")
        except Exception as ex:
            log.exception(ex)
        else:
            log.exception(f"No message in topic: {topico}")

        return mensajes

    def envia_mensaje(self, topico, mensaje):
        try:
            self.kafka_obj.send_message(topico, mensaje)
        except Exception as ex:
            log.exception(ex)




config_params = {
    'database': [
        {
            'host': 'ORACLE_HOST'
        },
        {
            'port': 'ORACLE_PORT'
        },
        {
            'user': 'ORACLE_USER'
        },
        {
            'pwd': 'ORACLE_PWD'
        },
        {
            'service': 'ORACLE_DB'
        },
        {
            'sid': 'ORACLE_SID'
        },
        {
            'schema': 'ORACLE_SCHEMA'
        }
    ],
    'kafka': [
        {
            'host': 'KAFKA_SERVER'
        },
        {
            'port': 'KAFKA_PORT'
        },
        {
            'user': 'KAFKA_USER'
        },
        {
            'pwd': 'KAFKA_PWD'
        }
    ],
    'sftp': [
        {
            'host': 'SFTP_HOST'
        },
        {
            'port': 'SFTP_PORT'
        },
        {
            'user': 'SFTP_USER'
        },
        {
            'pwd': 'SFTP_PWD'
        }
    ]
}



import json
from datetime import datetime

import chardet
import pymongo
from pymongo import MongoClient

from .mongo_collections import MONGO_COLLECTION_PROC_RECONCILED_CNF, MONGO_COLLECTIONS, \
    MONGO_COLLECTION_DATA_CONFIG_PARAMS
from .mongo_settings import MONGO_DB_USER, MONGO_DB_NAME, MONGO_DB_PORT, MONGO_DB_HOST, MONGO_DB_PWD, MONGO_DB_URI
from ..config import constantes

# Configuración del cursor para obtener los datos en lotes
size_lote = constantes.MONGO_BATCH_SIZE  # Número de documentos por lote


class MongoAPI:
    def __init__(self, data=None):
        self.total_data = 0
        self.offset = 20
        self.limit = 10

        self.starting_id = ''
        self.last_id = ''

        if MONGO_DB_USER == '':
            self.client = MongoClient(
                MONGO_DB_URI,
                compressors='zlib',
                directConnection=True
            )
        else:
            self.client = MongoClient(
                host=f'{MONGO_DB_HOST}:{MONGO_DB_PORT}',
                username=MONGO_DB_USER,
                password=MONGO_DB_PWD,
                # replicaset='dbrs',
                compressors='zlib',
                directConnection=True
            )

        if MONGO_DB_NAME not in self.client.list_database_names():
            self.db = self.client[f'{MONGO_DB_NAME}']
        else:
            self.db = self.client.get_database(f'{MONGO_DB_NAME}')

        if data:
            if 'collection' in data:
                self.collection = self.db[data['collection']]
            self.data = data

    def check_collections(self, custom_collections=None):
        collections = custom_collections if custom_collections else MONGO_COLLECTIONS
        for collection in collections:
            if collection not in self.db.list_collection_names():
                self.db.create_collection(collection)

    def read(self):
        """
        Retorna el documento disponible en base a filtro, ordenado y/o agrupado
        :return:
        """
        return self.__search_documents(
            self.data['Filter'] if 'Filter' in self.data else None,
            self.data['Sort'] if 'Sort' in self.data else None,
            self.data['Group'] if 'Group' in self.data else None,
            self.data['Columns'] if 'Columns' in self.data else None
        )

    def all(self):
        """
        Retorna todos los documentos disponibles a ser procesados basándose en filtros, ordenado y/o agrupado
        :return:
        """
        return self.__search_documents(
            self.data['Filter'] if 'Filter' in self.data else None,
            self.data['Sort'] if 'Sort' in self.data else None,
            self.data['Group'] if 'Group' in self.data else None,
            self.data['Columns'] if 'Columns' in self.data else None
        )

    def write(self, data, force_new=False):
        """
        Función que permite insertar un nuevo Documento en caso de no existir en la BD Mongo
        :param data:
        :param force_new
        :return: list[Mapping[str, Any]]
        """
        output = self.read()
        if output and not force_new:
            document_id = str(output[0]['_id'])
            self.update()
        else:
            new_document = data['Document']
            response = self.collection.insert_one(new_document)
            document_id = str(response.inserted_id)

        output = {'Status': 'Creación exitosa', 'Document_ID': document_id}
        return output

    def write_many(self, documents):
        """
        Función que permite insertar un nuevo Documento en caso de no existir en la BD Mongo
        :param documents:
        :return: list[Mapping[str, Any]]
        """
        output = self.read()
        if output:
            document_id = str(output[0]['_id'])
        else:
            response = self.collection.insert_many(documents)
            document_id = str(response.inserted_ids)

        output = {'Status': 'Creación exitosa', 'Document_ID': document_id}
        return output

    def update(self):
        """
        Función que actualiza la información en la BD Mongo
        :return: dict[str, str]
        """
        if 'DataToBeUpdated' in self.data:
            _filter = self.data['Filter']
            updated_data = {"$set": self.data['DataToBeUpdated']}
            response = self.collection.update_one(_filter, updated_data)
            output = {
                'Status': 'Actualización exitosa' if response.modified_count > 0 else "No hay nada que actualizar."}
            return output
        return None

    def delete(self, data):
        """
        Elimina lógicamente un registro desde la BD Mongo
        :param data:
        :return:
        """
        _filter = data['Filter']
        response = self.collection.delete_one(_filter)
        output = {'Status': 'Eliminación exitosa' if response.deleted_count > 0 else "Documento no encontrado."}
        return output

    def delete_many(self, data):
        """
        Elimina lógicamente un registro desde la BD Mongo
        :param data:
        :return:
        """
        _filter = data['Filter']
        response = self.collection.delete_many(_filter)
        output = {'Status': 'Eliminación exitosa' if response.deleted_count > 0 else "Documento no encontrado."}
        return output

    def read_with_pagination(self, data):
        _filter = data['Filter']
        output = []
        self.total_data = self.collection.count_documents(_filter)
        if self.total_data > 0:
            self.offset = int(data['Pagination']['offset'])  # corresponde al item que inicia el resultado obtenido
            if self.offset > self.total_data:
                self.offset = self.total_data

            self.limit = int(data['Pagination']['limit'])  # corresponde a la cantidad de datos devueltos
            self.starting_id = self.collection.find(_filter).sort('_id', pymongo.ASCENDING)
            self.last_id = self.starting_id[self.offset - 1]['_id']

            documents = self.collection.find({"$and": [_filter, {'_id': {'$gte': self.last_id}}]}) \
                .sort('_id', pymongo.ASCENDING) \
                .limit(self.limit)

            for document in documents:
                document['_id'] = str(document['_id'])
                del document[f'{MONGO_COLLECTION_PROC_RECONCILED_CNF}_id']
                output.append(document)

        next_offset = self.offset + self.limit
        prev_offset = self.offset - self.limit
        return {
            'next': next_offset,
            'prev': prev_offset if prev_offset > 0 else 1,
            'data': output,
            'total': self.total_data
        }

    def __search_documents(self, _filter=None, _sort=None, _group=None, _cols=None):
        if _filter is None:
            _filter = {}
        else:
            columns_filter = ''
            init = True
            for key, value in _filter.items():
                if init:
                    columns_filter = f"{key}"
                    init = False
                else:
                    columns_filter += f",{key}"
            self.collection.create_index(columns_filter)

        output = []

        if _group:
            self.collection.aggregate(
                [{
                    '$group': _group
                }]
            )

        lote_actual = 0
        total_datos = self.collection.count_documents(_filter)
        while lote_actual < total_datos:
            if _sort:
                lote = self.collection.find(_filter, _cols if _cols else {}).skip(lote_actual).limit(size_lote).sort(_sort)
            else:
                lote = self.collection.find(_filter, _cols if _cols else {}).skip(lote_actual).limit(size_lote)

            for documento in lote:
                documento['_id'] = str(documento['_id'])
                if self.data['collection'] != MONGO_COLLECTION_DATA_CONFIG_PARAMS:
                    for param in ['created_at', 'deleted_at']:
                        if param in documento:
                            documento = json.loads(json.dumps(documento, cls=DateTimeEncoder, ensure_ascii=False))
                output.append(documento)

            lote_actual += size_lote
        return output


class DateTimeEncoder(json.JSONEncoder):
    def default(self, z):
        if isinstance(z, datetime):
            return str(z)
        else:
            return super().default(z)



from Crypto.Cipher import AES
from Crypto.Random import get_random_bytes

from ..config import constantes
from ..config.config_params import config_params
from ..database import MongoAPI
from ..database.mongo_collections import MONGO_COLLECTION_DATA_CONFIG_PARAMS
from ..enums.enums import EventsLogsEnum
from ..log import AuditoryLogs
from ..utility import get_datetime


def __init_set_aes_key():
    aes_key = get_random_bytes(16)
    data = {
        'collection': MONGO_COLLECTION_DATA_CONFIG_PARAMS,
        'Document': {
            'name_param': 'aes_key',
            'value_param': aes_key,
            'created_at': get_datetime()
        },
        'Filter': {
            'name_param': 'aes_key'
        }
    }
    mongodb = MongoAPI(data)
    mongodb.write(data)
    return aes_key


def set_config_params(json_dict: dict):
    data = {
        'collection': MONGO_COLLECTION_DATA_CONFIG_PARAMS,
        'Filter': {
            'name_param': 'aes_key'
        }
    }
    mongodb = MongoAPI(data)
    response = mongodb.read()
    if len(response) == 0:
        aes_key = __init_set_aes_key()
    else:
        aes_key = response[0]['value_param']

    cipher = AES.new(aes_key, AES.MODE_OCB)
    ciphertext, tag = cipher.encrypt_and_digest(json_dict.get('value_param').encode())

    response = None
    try:
        data = {
            'collection': MONGO_COLLECTION_DATA_CONFIG_PARAMS,
            'Document': {
                'type_param': json_dict.get('type_param'),
                'name_param': json_dict.get('name_param'),
                'tag_param': tag,
                'nonce_param': cipher.nonce,
                'text_param': ciphertext,
                'created_date': get_datetime()
            },
            'Filter': {
                'type_param': json_dict.get('type_param'),
                'name_param': json_dict.get('name_param')
            },
            'DataToBeUpdated': {
                'tag_param': tag,
                'nonce_param': cipher.nonce,
                'text_param': ciphertext,
                'updated_date': get_datetime()
            }
        }

        mongodb = MongoAPI(data)
        response = mongodb.write(data)
    except Exception as ex:
        AuditoryLogs.registry_log(
            origin=f'{__name__}.set_config_params',
            event=EventsLogsEnum.EVENT_ERROR,
            description=f'Error al obtener los datos: {ex}',
            user=constantes.USER_DEFAULT
        )

    return response


def get_config_params(p, t):
    json_dict = {
        'name_param': p,
        'type_param': t
    }

    response = None
    data = {
        'collection': MONGO_COLLECTION_DATA_CONFIG_PARAMS,
        'Filter': {
            'name_param': 'aes_key'
        }
    }
    mongodb = MongoAPI(data)
    result = mongodb.read()
    if len(result) == 0:
        aes_key = __init_set_aes_key()
    else:
        aes_key = result[0]['value_param']

    try:
        data = {
            'collection': MONGO_COLLECTION_DATA_CONFIG_PARAMS,
            'Filter': {
                'name_param': json_dict.get('name_param'),
                'type_param': json_dict.get('type_param')
            }
        }

        mongodb = MongoAPI(data)
        response = mongodb.read()
        if len(response) > 0:
            key = dict(response[0])
            ciphertext = key.get('text_param')
            tag = key.get('tag_param')
            nonce = key.get('nonce_param')

            cipher = AES.new(aes_key, AES.MODE_OCB, nonce=nonce)
            try:
                message = cipher.decrypt_and_verify(ciphertext, tag)
                response = message.decode()
            except ValueError as ex:
                AuditoryLogs.registry_log(
                    origin=f'{__name__}.get_config_params',
                    event=EventsLogsEnum.EVENT_ERROR,
                    description=f'Error al obtener los datos: {ex}',
                    user=constantes.USER_DEFAULT
                )
        else:
            AuditoryLogs.registry_log(
                origin=f'{__name__}.get_config_params',
                event=EventsLogsEnum.EVENT_ERROR,
                description=f'El parámetro {json_dict.get("name_param")} no existe',
                user=constantes.USER_DEFAULT
            )

    except Exception as ex:
        AuditoryLogs.registry_log(
            origin=f'{__name__}.get_config_params',
            event=EventsLogsEnum.EVENT_ERROR,
            description=f'Error al obtener los datos: {ex}',
            user=constantes.USER_DEFAULT
        )

    return response


def get_all_config_params_by_type(t):
    response_data = None
    try:
        data = {
            'collection': MONGO_COLLECTION_DATA_CONFIG_PARAMS,
            'Filter': {
                'name_param': 'aes_key'
            }
        }
        mongodb = MongoAPI(data)
        result = mongodb.read()
        if len(result) == 0:
            aes_key = __init_set_aes_key()
        else:
            aes_key = result[0]['value_param']

        data_type = config_params.get(t)
        if len(data_type) > 0:
            data = {
                'collection': MONGO_COLLECTION_DATA_CONFIG_PARAMS,
                'Filter': {
                    'type_param': t
                }
            }

            mongodb = MongoAPI(data)
            response = mongodb.read()
            if len(response) > 0:
                response_data = {
                    'conf': t,
                    'data': {}
                }
                for key in response:
                    ciphertext = key.get('text_param')
                    tag = key.get('tag_param')
                    nonce = key.get('nonce_param')

                    cipher = AES.new(aes_key, AES.MODE_OCB, nonce=nonce)
                    try:
                        dict_response = [d for d in data_type if key.get('name_param') in d.values()][0]

                        key_list = list(dict_response.keys())
                        val_list = list(dict_response.values())
                        position = val_list.index(key.get('name_param'))

                        message = cipher.decrypt_and_verify(ciphertext, tag)
                        response_data['data'][f'{key_list[position]}'] = message.decode()
                    except ValueError as ex:
                        AuditoryLogs.registry_log(
                            origin=f'{__name__}.get_all_config_params_by_type',
                            event=EventsLogsEnum.EVENT_ERROR,
                            description=f'Error al obtener los datos: {ex}',
                            user=constantes.USER_DEFAULT
                        )
            else:
                response_data = []
                AuditoryLogs.registry_log(
                    origin=f'{__name__}.get_all_config_params_by_type',
                    event=EventsLogsEnum.EVENT_ERROR,
                    description=f'El tipo parámetro {t} no existe',
                    user=constantes.USER_DEFAULT
                )
        else:
            AuditoryLogs.registry_log(
                origin=f'{__name__}.get_all_config_params_by_type',
                event=EventsLogsEnum.EVENT_ERROR,
                description=f'El tipo parámetro {t} no existe',
                user=constantes.USER_DEFAULT
            )

    except Exception as ex:
        AuditoryLogs.registry_log(
            origin=f'{__name__}.get_all_config_params_by_type',
            event=EventsLogsEnum.EVENT_ERROR,
            description=f'Error al obtener los datos: {ex}',
            user=constantes.USER_DEFAULT
        )

    return response_data



# MongoDB Documents
MONGO_COLLECTION_UPLOAD_FILE_USER_ID = 'user_id'
MONGO_COLLECTION_UPLOAD_FILE_PARTNER_ID = 'partner_id'
MONGO_COLLECTION_UPLOAD_FILE_CLIENT_ID = 'partner_id'
MONGO_COLLECTION_UPLOAD_FILE = 'upload_files'
MONGO_COLLECTION_CONFIG_RULES = 'config_rules'
MONGO_COLLECTION_CONFIG_RULES_DEFINED_COLUMNS = 'config_rules_columns'
MONGO_COLLECTION_CONFIG_RULES_DEFINED_INDEX_COLUMNS = 'config_rules_columns_index'
MONGO_COLLECTION_CONFIG_RULES_DEFINED_AMOUNT_COLUMN = 'config_rules_column_amount'
MONGO_COLLECTION_CONFIG_RULES_DEFINED_DATE_COLUMN = 'config_rules_column_date'
MONGO_COLLECTION_CONFIG_RULES_DEFINED_MAPPING_COLUMNS = 'config_rules_columns_mapping'
MONGO_COLLECTION_PROC_RECONCILED_CNF = 'config_process'
MONGO_COLLECTION_PROC_RECONCILED_FILE = 'config_process_files'
MONGO_COLLECTION_DATA_RECONCILED = 'reconciled'
MONGO_COLLECTION_DATA_TOTAL_RECONCILED = 'total_reconciled'
MONGO_COLLECTION_DATA_DUPLICATES = 'duplicates'
MONGO_COLLECTION_DATA_TOTAL_DUPLICATES = 'total_duplicates'
MONGO_COLLECTION_DATA_UNRECONCILED = 'unreconciled'
MONGO_COLLECTION_DATA_TOTAL_UNRECONCILED = 'total_unreconciled'
MONGO_COLLECTION_DATA_USERS = 'users'
MONGO_COLLECTION_DATA_ROLES = 'roles'
MONGO_COLLECTION_DATA_PARTNERS = 'partners'
MONGO_COLLECTION_DATA_TYPE_CYCLES = 'type_cycles'
MONGO_COLLECTION_DATA_CYCLES = 'cycles'
MONGO_COLLECTION_DATA_PROG_SCHEDULERS = 'program_scheduler'
MONGO_COLLECTION_AUDITORY_LOGS = 'auditory_logs'
MONGO_COLLECTION_DATA_CONFIG_PARAMS = 'config_params'

MONGO_COLLECTIONS = {
    MONGO_COLLECTION_DATA_RECONCILED,
    MONGO_COLLECTION_DATA_TOTAL_RECONCILED,
    MONGO_COLLECTION_DATA_UNRECONCILED,
    MONGO_COLLECTION_DATA_TOTAL_UNRECONCILED,
    MONGO_COLLECTION_DATA_DUPLICATES,
    MONGO_COLLECTION_DATA_TOTAL_DUPLICATES,
    MONGO_COLLECTION_CONFIG_RULES,
    MONGO_COLLECTION_CONFIG_RULES_DEFINED_COLUMNS,
    MONGO_COLLECTION_CONFIG_RULES_DEFINED_INDEX_COLUMNS,
    MONGO_COLLECTION_CONFIG_RULES_DEFINED_AMOUNT_COLUMN,
    MONGO_COLLECTION_CONFIG_RULES_DEFINED_DATE_COLUMN,
    MONGO_COLLECTION_CONFIG_RULES_DEFINED_MAPPING_COLUMNS,
    MONGO_COLLECTION_UPLOAD_FILE,
    MONGO_COLLECTION_PROC_RECONCILED_CNF,
    MONGO_COLLECTION_PROC_RECONCILED_FILE,
    MONGO_COLLECTION_DATA_USERS,
    MONGO_COLLECTION_DATA_ROLES,
    MONGO_COLLECTION_DATA_PARTNERS,
    MONGO_COLLECTION_DATA_TYPE_CYCLES,
    MONGO_COLLECTION_DATA_CYCLES,
    MONGO_COLLECTION_DATA_PROG_SCHEDULERS,
    MONGO_COLLECTION_AUDITORY_LOGS,
    MONGO_COLLECTION_DATA_CONFIG_PARAMS
}


import logging

from ..config import settings, constantes
from ..enums.enums import EventsLogsEnum
from ..log import AuditoryLogs
from ..utility.utils_messages import UtilsMessages

log = logging.getLogger(__name__)


def petition(json_dict):
    try:
        UtilsMessages.send_messages(settings.KAFKA_TOPIC_PETITION, json_dict)
    except Exception as ex:
        log.exception(ex)
        AuditoryLogs.registry_log(
            origin=f'{__name__}.petition',
            event=EventsLogsEnum.EVENT_ERROR,
            description=f'Error al almacenar los datos: {ex}',
            user=constantes.USER_DEFAULT
        )


def scheduler(json_dict):
    try:
        UtilsMessages.send_messages(settings.KAFKA_TOPIC_SCHEDULER, json_dict)
    except Exception as ex:
        log.exception(ex)
        AuditoryLogs.registry_log(
            origin=f'{__name__}.scheduler',
            event=EventsLogsEnum.EVENT_ERROR,
            description=f'Error al almacenar los datos: {ex}',
            user=constantes.USER_DEFAULT
        )



import datetime
import logging

from bson import ObjectId

from ..config import settings
from ..database.mongo_collections import MONGO_COLLECTION_DATA_ROLES, MONGO_COLLECTION_DATA_PARTNERS
from ..database.mongodb import MongoAPI
from ..utility.utils import get_message_error, get_datetime

log = logging.getLogger(__name__)


def new_role(data):
    try:
        date_doc = datetime.datetime.now()
        data = {
            'collection': MONGO_COLLECTION_DATA_ROLES,
            'Document': {
                'description': data.get('description'),
                'is_admin': (data.get('description') == f'{settings.DEFAULT_ROLE_NAME_ADMIN}'),
                'active': data.get('active'),
                'created_at': date_doc
            },
            'Filter': {
                'description': data.get('description')
            },
            'DataToBeUpdated': {
                'description': data.get('description'),
                'updated_at': date_doc,
            }
        }

        mongodb = MongoAPI(data)
        response = mongodb.write(data)

    except Exception as e:
        log.exception(e)
        response = get_message_error(e)
    return response


def get_role(role_id):
    try:
        data = {
            'collection': MONGO_COLLECTION_DATA_ROLES,
            'Filter': {
                '_id': ObjectId(role_id)
            }
        }
        mongodb = MongoAPI(data)
        response = mongodb.read()
    except Exception as e:
        log.exception(e)
        response = get_message_error(e)
    return response


def get_all_roles():
    try:
        data = {
            'collection': MONGO_COLLECTION_DATA_ROLES,
            'Filter': {
                'active': True,
                'is_admin': False
            }
        }
        mongodb = MongoAPI(data)
        response = mongodb.all()
    except Exception as e:
        log.exception(e)
        response = get_message_error(e)
    return response


def update_role(pid: str, request):
    try:
        data = {
            'collection': MONGO_COLLECTION_DATA_PARTNERS,
            'Filter': {
                '_id': ObjectId(pid)
            },
            'DataToBeUpdated': {
                'description': request.get('nombre'),
                'active': request.get('activo'),
                'updated_at': get_datetime(),
            }
        }
        mongodb = MongoAPI(data)
        response = mongodb.update()
    except Exception as e:
        log.exception(e)
        response = get_message_error(e)
    return response


def delete_role(pid: str):
    try:
        data = {
            'collection': MONGO_COLLECTION_DATA_ROLES,
            'Filter': {
                '_id': ObjectId(pid)
            }
        }
        mongodb = MongoAPI(data)
        response = mongodb.write(data)

        if response:
            data = {
                'collection': MONGO_COLLECTION_DATA_ROLES,
                'Filter': {
                    '_id': ObjectId(pid)
                },
                'DataToBeUpdated': {
                    'active': False,
                    'updated_at': get_datetime(),
                }
            }
            mongodb = MongoAPI(data)
            response = mongodb.update()
            if response:
                return {'Status': 'Eliminación exitosa'}

    except Exception as e:
        log.exception(e)
        response = get_message_error(e)
    return response


def check_role_exists(description):
    try:
        data = {
            'collection': MONGO_COLLECTION_DATA_ROLES,
            'Filter': {
                'description': description
            }
        }
        mongodb = MongoAPI(data)
        response = mongodb.read()
    except Exception as e:
        log.exception(e)
        response = get_message_error(e)
    return response



RESPONSE_ARGUMENT_INVALID = 'Argumento inválido.'
RESPONSE_MAPPING_KEY_ERROR = 'Error en parámetro ingresado'
RESPONSE_NOT_AUTHORIZED = 'No autorizado'
RESPONSE_SUCCESS = 'Respuesta obtenida exitosamente.'
RESPONSE_ERROR = 'Ah ocurrido un error al tratar de obtener la respuesta del servicio.'
RESPONSE_CREATE_SUCCESS = 'Creación exitosa'
RESPONSE_CREATE_ERROR = 'Ah ocurrido un error al tratar de crear el registro.'
RESPONSE_UPDATE_SUCCESS = 'Actualización exitosa'
RESPONSE_UPDATE_ERROR = 'Ah ocurrido un error al tratar de actualizar el registro.'
RESPONSE_DELETE_SUCCESS = 'Eliminación exitosa'
RESPONSE_DELETE_ERROR = 'Ah ocurrido un error al tratar de eliminar el registro.'
RESPONSE_VALID_TOKEN_IS_MISSING = 'No se tiene un token válido.'
RESPONSE_TOKEN_IS_INVALID = 'El token es inválido'
RESPONSE_DATA_EXISTS = 'Los datos ingresados ya existen'
RESPONSE_DATA_NOT_EXISTS = 'Los datos ingresados no existen'
RESPONSE_SUCCESS_UPLOAD = 'Archivo subido exitosamente.'
RESPONSE_ERROR_UPLOAD = 'Ah ocurrido un error al tratar de subir el archivo'
RESPONSE_NOT_SELECT_FILE_UPLOAD = 'No ha seleccionado un archivo para subir'
RESPONSE_EMPTY_SELECT_FILE_UPLOAD = 'El archivo seleccionado está vacío'
RESPONSE_NOT_PART_FILE_UPLOAD = 'No hay parte de archivo en la solicitud'
RESPONSE_NOT_SETTING_TYPE_FILE_UPLOAD = 'No ha indicado el parámetro type en la solicitud'
RESPONSE_EMPTY_SETTING_TYPE_FILE_UPLOAD = 'El valor del parámetro type non puede ser nulo o vacío'

# Login
LOGIN_API_DESCRIPTION = 'Acceso al Sistema'
LOGIN_API_ENDPOINT = 'login'
LOGIN_ERROR = 'No se puede verificar'
RESPONSE_LOGIN_INVALID = 'Login inválido'
RESPONSE_LOGIN_USER_PWD_INVALID = 'Usuario y Password incorrectos.'

# Register
REGISTER_API_DESCRIPTION = 'Registro de Usuario'
REGISTER_API_ENDPOINT = 'register'
REGISTER_ERROR = 'No se puede verificar'

# Rol
ROLE_API_DESCRIPTION = 'Administración de Roles'
ROLE_API_ENDPOINT = 'auth/role'
ROLE_DESCRIPTION_ID = 'Identificador del Rol'

# Usuario
USER_API_DESCRIPTION = 'Administración de Usuarios'
USER_API_ENDPOINT = 'auth/user'
USER_DESCRIPTION_ID = 'Identificador del Usuario'

# Socio
PARTNER_API_DESCRIPTION = 'Administración de Socios'
PARTNER_API_ENDPOINT = 'auth/partner'
PARTNER_DESCRIPTION_ID = 'Identificador del Socio'

# Conciliación
CONCILIATION_API_DESCRIPTION = 'Conciliación'
CONCILIATION_API_ENDPOINT = 'auth/conciliation'
CONCILIATION_DESCRIPTION_ID = 'Identificador de la Conciliación'

# Tipos de Ciclo
TYPE_CYCLES_API_DESCRIPTION = 'Administración de Tipos de Ciclo'
TYPE_CYCLES_API_ENDPOINT = 'auth/type-cycles'
TYPE_CYCLES_DESCRIPTION_ID = 'Identificador del Tipo Ciclo'

# Ciclo
CYCLES_API_DESCRIPTION = 'Ciclo'
CYCLES_API_ENDPOINT = 'auth/cycles'
CYCLES_DESCRIPTION_ID = 'Identificador del Ciclo'

# Endpoints API
PETITION_API_ENDPOINT = 'auth/petitions'
PETITION_API_DESCRIPTION = 'Peticiones Kafka'

# Reglas Conciliación
CONFIG_RULES_API_ENDPOINT = 'auth/config'
CONFIG_RULES_API_DESCRIPTION = 'Configuración de Reglas para la Conciliación'
CONFIG_RULES_DESCRIPTION_ID = 'Identificador de la Configuración'
CONFIG_RULES_YEAR = 'Año asociado al periodo a ejecutar la Conciliación'
CONFIG_RULES_MONTH = 'Mes asociado al periodo a ejecutar la Conciliación'

# Estado de la configuración de la Conciliación
STATE_CONFIG_PENDING = 'Creada'
STATE_CONFIG_RECONCILED = 'Configurada'
STATE_CONFIG_EXECUTED = 'Ejecutada'
STATE_CONFIG_IN_PROCESS = 'En Ejecución'
STATE_CONFIG_NOT_PROCESSED = 'No Procesada'
STATE_CONFIG_SCHEDULER = 'Programada'

CODE_STATE_CONFIG_PENDING = 'pending'
CODE_STATE_CONFIG_RECONCILED = 'config'
CODE_STATE_CONFIG_EXECUTED = 'executed'
CODE_STATE_CONFIG_IN_PROCESS = 'in-process'
CODE_STATE_CONFIG_NOT_PROCESSED = 'not-processed'
CODE_STATE_CONFIG_SCHEDULER = 'scheduler'

# Configuración Aplicación
CONFIG_APP_API_ENDPOINT = 'auth/config_app'
CONFIG_APP_API_DESCRIPTION = 'Configuración de parámetros de la Aplicación'
CONFIG_APP_API_DESCRIPTION_ID = 'Identificador del parámetro a consultar'

# Configuración Aplicación
SCHEDULER_API_ENDPOINT = 'auth/task'
SCHEDULER_API_DESCRIPTION = 'Configuración de Tareas de la Aplicación'
SCHEDULER_API_DESCRIPTION_ID = 'Identificador del parámetro a consultar'

# Opciones de Reportes
OPTIONS_REPORT = 'Opción para emisión de Reportes (Conciliación / No Conciliación / Duplicados)'
OPTIONS_LIMIT = 'Corresponde a la cantidad de datos a entregar'
OPTIONS_OFFSET = 'Corresponde a el valor desde donde se inicia la búsqueda del item en la consulta'
OPTIONS_REPORT_OPC = ('Corresponde a que reporte se desea obtener'
                      ', los que pueden ser Conciliados, No Conciliados, Duplicados o Todos')
OPTIONS_ACT = 'Corresponde a la pagina actual'
OPTIONS_TYPE = 'Corresponde al tipo de reporte a descargar (CSV, XLSX)'
OPTIONS_TYPE_PROCESS = 'Corresponde al tipo de proceso del reporte (Primario / Externo)'
OPTION_REPORTS_ALL = 'all'
OPTION_REPORTS_RECONCILED = 'reconciled'
OPTION_REPORTS_UNRECONCILED = 'unreconciled'
OPTION_REPORTS_DUPLICATED = 'duplicated'

# Constantes Globales
EMPTY = ''
H5_FILE = '.h5'
MONGO_BATCH_SIZE = 100000

# CORS Configuration
CORS_ALLOW_ORIGIN = "*,*"
CORS_EXPOSE_HEADERS = "*,*"
CORS_ALLOW_HEADERS = "content-type,Access-Control-Allow-Origin,*"

USER_DEFAULT = 'optidata-core'



from .config import (
    api_rest,
    constantes,
    error_handling,
    kafka_config,
    settings,
    sftp
)
from .database import (
    mongodb,
    oracle
)
from .enums import enums
from .log import auditory
from .messages import messages
from .services import (
    conciliation,
    config_rules,
    cycles,
    partner,
    role,
    type_cycles,
    users,
    messages
)
from .utility import (
    utils,
    utils_file,
    utils_messages
)



import calendar
import json
import logging
from datetime import datetime

from bson import ObjectId
from flask import jsonify, make_response
from flask_bcrypt import generate_password_hash, check_password_hash

from ..config import settings

log = logging.getLogger(__name__)


def empty_result():
    return []


def set_default(obj):
    if isinstance(obj, set):
        return list(obj)
    raise TypeError


def cast_id_mongo(id_mongo):
    return str(id_mongo)


def get_uuid():
    return cast_id_mongo(ObjectId())


def get_datetime():
    date_new_format = "%d-%m-%Y %H:%M:%S"
    return datetime.now().strftime(date_new_format)


def allowed_file(filename):
    return '.' in filename and filename.rsplit('.', 1)[1].lower() in settings.FLASK_ALLOWED_EXTENSIONS


def password(pwd: str):
    log_rounds = settings.BCRYPT_LOG_ROUNDS
    hash_bytes = generate_password_hash(pwd, log_rounds)
    return hash_bytes.decode("utf-8")


def check_password(pwd_hash, pwd):
    return check_password_hash(pwd_hash, pwd)


def get_mime_type_application(type_extension: str):
    if type_extension == 'csv':
        return "text/csv"
    elif type_extension == 'xls' or type_extension == 'xlsx':
        return "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"


def get_message_error(err):
    return make_response(jsonify({
        "msg": f"Ah ocurrido un error: {err}"
    }), 400)


def get_message_success(msg):
    return make_response(jsonify({
        "msg": f"{msg}"
    }), 200)


def get_start_end_day_month(month, year):
    start_day = 1
    end_day = calendar.monthrange(year, month)[1]
    return start_day, end_day


# Messages will be serialized as JSON
def serializer(message):
    return json.dumps(message).encode('utf-8')


def extract_data_line(data: str, pos_ini=None, total=None):
    if pos_ini and total:
        return data[pos_ini:total]
    elif pos_ini:
        return data[pos_ini:]
    else:
        return data[:total]



import logging
import os
import time

import numpy as np
# from numba import cuda, typed
import pandas as pd
import vaex
from joblib import Parallel, delayed
from pandas.errors import InvalidIndexError

from ..config import constantes
from ..config import settings
from ..enums.enums import DataTypeEnum
from ..utility.utils import get_start_end_day_month

# import cudf
# import torch

# Configura las opciones de Pandas para mejorar el rendimiento
pd.options.mode.chained_assignment = None  # Deshabilita las advertencias de asignaciones
chunk_size: int = 100000  # Tamaño de cada fragmento de lectura

# Item en donde se produce el error en el proceso de Consolidación
item_error = ''

# Establecer el formato de moneda local
# locale.setlocale(locale.LC_ALL, 'es_CL.UTF-8')

log = logging.getLogger(__name__)


# Funciones Privadas
def __getfile_from_path(sub_path: str, filename: str, ext: str):
    return os.path.join(settings.FLASK_UPLOAD_FOLDER, sub_path, f'{filename}{ext}')


# @cuda.jit(target="cuda")
def __get_columns_dataframe(file, ext: str):
    cols = None
    if ext == f'.{settings.FILE_EXT_TXT}':
        pass
    elif ext == f'.{settings.FILE_EXT_CSV}':
        cols = pd.read_csv(file, sep=settings.CSV_SEPARATOR).to_dict()
    elif ext == f'.{settings.FILE_EXT_XLS}' or ext == f'.{settings.FILE_EXT_XLSX}':
        cols = pd.read_excel(file.read(), nrows=1).to_dict()
    else:
        pass

    return __remove_spaces_name(cols)


# @cuda.jit(target="cuda")
def __get_dataframe(ext: str, file_name: str, path_file: str):
    df = pd.DataFrame()
    if ext == f'.{settings.FILE_EXT_TXT}':
        pass
    elif ext == f'.{settings.FILE_EXT_CSV}':
        df = pd.read_csv(path_file, sep=settings.CSV_SEPARATOR, dayfirst=True)
    elif ext == f'.{settings.FILE_EXT_XLS}' or ext == f'.{settings.FILE_EXT_XLSX}':
        dfs = __read_excel(path_file)
        df = pd.concat(dfs, ignore_index=True)
    else:
        df = vaex.open(path=f'{path_file}', group=f'{file_name}').to_pandas_df()

    return df


def __read_excel(path_file: str):
    xls = pd.ExcelFile(path_file)
    nombres_hojas = xls.sheet_names
    dfs = Parallel(n_jobs=-1)(delayed(__loop_read_excel)(path_file, hoja) for hoja in nombres_hojas)
    return dfs


# @cuda.jit(target="cuda")
def __loop_read_excel(file, sheet_name):
    return pd.read_excel(file, sheet_name=sheet_name)


# @cuda.jit(target="cuda")
def __normalized_data(df, key: str, type_data: str):
    if type_data == DataTypeEnum.DATE.value:
        df[f'{key}'] = pd.to_datetime(df[f'{key}'], format='mixed')
    elif type_data == DataTypeEnum.STRING.value:
        df[f'{key}'] = df[f'{key}'].fillna('-').astype(str).apply(lambda x: x.strip())
    elif type_data == DataTypeEnum.INTEGER.value:
        df[f'{key}'] = df[f'{key}'].astype(float).astype('Int64').fillna(0)
    elif type_data == DataTypeEnum.FLOAT.value:
        df[f'{key}'] = df[f'{key}'].astype(float).astype('Int64').fillna(0)
    else:
        df[f'{key}'] = df[f'{key}'].fillna('-').astype(str).apply(lambda x: x.strip())

    return df[f'{key}']


def __remove_spaces_name_columns(df):
    trim_name_columns = []
    for col in df.columns:
        trim_name_columns.append(col.strip())
    df.columns = trim_name_columns
    return df


def __remove_spaces_name(cols):
    trim_name_columns = []
    for col in cols:
        trim_name_columns.append(col.strip())
    return trim_name_columns


def __change_data_nat_nan_to_str(df):
    for col in df.columns:
        if df[col].dtype == np.dtype('object'):
            df[col] = df[col].where(df[col].notnull(), '')
            df[col] = df[col].where(df[col].notna(), '')
            df[col] = df[col].astype("string")
        elif df[col].dtype == np.dtype('datetime64[ns]'):
            df[col] = df[col].astype(object).where(df[col].notnull(), '')
            df[col] = df[col].astype(object).where(df[col].notna(), '')
            df[col] = df[col].astype("string")
        elif df[col].dtype == np.dtype('float64') or df[col].dtype == np.dtype('int64'):
            df[col] = df[col].where(df[col].notnull(), 0)
            df[col] = df[col].where(df[col].notna(), 0)
            df[col] = df[col].astype("int64")

    return df


def __create_index_in_df(df, index_key: str, cols_list: list):
    init = False
    for col in cols_list:
        if not init:
            df[f'{index_key}'] = df[f'{col}'].fillna(0).astype(str).apply(lambda x: x.strip())
            init = True
        else:
            df[f'{index_key}'] = df[f'{index_key}'] + df[f'{col}'].fillna(0).astype(str).apply(lambda x: x.strip())

    if init:
        df.set_index(f'{index_key}', inplace=True)
    return df


def __sort_datetime_data(df, column_name: str):
    return df.sort_values(by=[f'{column_name}'])


def __convert_datetime_to_str(df, key: str, type_data: str):
    if type_data == DataTypeEnum.DATE.value:
        df[f'{key}'] = df[f'{key}'].dt.strftime('%Y-%m-%d %H:%M:%S')

    return df[f'{key}']


def __get_duplicate(df):
    return df.loc[df.duplicated(keep=False)]


def __delete_duplicate(df, columns):
    return df.drop_duplicates(columns)


def __concat_data(df_origin, df_compared, union_columns=False):
    df_origin = df_origin.loc[~df_origin.index.duplicated(keep='first')]
    df_compared = df_compared.loc[~df_compared.index.duplicated(keep='first')]
    return pd.concat([df_origin, df_compared], axis=union_columns)


# @cuda.jit(target="cuda")
def __merge_data(df_origin, df_compared, name_index, is_common=True):
    _suffixes = ('_primary', '_external')
    if is_common:
        df = pd.merge(
            df_origin,
            df_compared,
            on=name_index,
            suffixes=_suffixes,
            how='inner'
        )
    else:
        df = pd.merge(
            df_origin,
            df_compared,
            on=name_index,
            suffixes=_suffixes,
            how='outer',
            indicator=True
        ).query("_merge != 'both'")

    return df


def __get_totals_dataframe_by_column(df, key: str):
    return df[f'{key}'].sum()


def __get_data_filter_dates(
        df,
        column,
        start_day,
        end_day,
        month: int,
        year: int):
    if len(f'{month}') < 2:
        m = f'0{month}'
    else:
        m = month

    init_date = pd.to_datetime(f'{year}-{m}-0{start_day}').date()
    finish_date = pd.to_datetime(f'{year}-{m}-{end_day}').date()

    return df[
        (pd.to_datetime(df[f'{column}']).dt.date >= init_date) &
        (pd.to_datetime(df[f'{column}']).dt.date <= finish_date)
        ]


def __add_column_description(df_c, df_nc, idx_columns):
    msg_diff = ""
    msg_part = "en la columna"
    is_more = False
    for col in idx_columns:
        if not np.where(df_c[f'{col}_primary'] == df_nc[f'{col}']):
            if len(msg_diff) == 0:
                msg_diff = f"{col}"
            else:
                msg_diff += f", {col}"
                is_more = True

    if is_more:
        msg_part = "en las columnas"
    return f"Existe una diferencia {msg_part} {msg_diff}." if len(msg_diff) > 0 else ""


##########################################
# Functions Public's
##########################################
def extract_headers_file(file, ext: str):
    columns_list = []
    try:
        df_columns = __get_columns_dataframe(file, ext)

        if df_columns:
            for column in df_columns:
                columns_list.append({
                    "col_name": f"{column}",
                    "col_used": False,
                    "col_type": "",
                    "col_norm": False
                })

            return columns_list
    except Exception as e:
        logging.exception(e)

    return columns_list


def create_file_from_dataframe(df, name_file: str):
    df.to_csv(f"{settings.FLASK_EXPORT_FOLDER}/{name_file}.csv", index=True, sep=';', encoding='utf-8-sig')


# @cuda.jit(target="cuda")
def execute_process_conciliation(data: list, doc_id: str):
    global item_error
    result_process = {}
    try:
        data_item = data[0]
        # Definición de variables base
        columns_df = __remove_spaces_name(data_item['columns_df'])
        sub_path = os.path.join(
            f'{data_item["internal"]["partner"]}',
            f'{data_item["internal"]["year"]}',
            f'{data_item["internal"]["month"]}'
        )

        # Carga de datos
        # Proceso Primary v/s External
        file_name = f"{data_item['internal']['filename']}{data_item['internal']['extension']}"
        item_error = f"Error lectura archivo: {file_name}"
        df_original = __get_dataframe(
            constantes.H5_FILE,  # data_item['internal']['extension'],
            data_item['internal']['filename'],
            __getfile_from_path(sub_path, data_item['internal']['filename'], constantes.H5_FILE)
        )

        file_name = f"{data_item['external'][0]['filename']}{data_item['external'][0]['extension']}"
        item_error = f"Error lectura archivo: {file_name}"
        df_compared = __get_dataframe(
            constantes.H5_FILE,  # data_item['external'][0]['extension'],
            data_item['external'][0]['filename'],
            __getfile_from_path(sub_path, data_item['external'][0]['filename'], constantes.H5_FILE)
        )

        # Obtiene fechas para proceso de filtrado de data para ejecución de consolidación de datos
        start_day, end_day = get_start_end_day_month(
            int(data_item["internal"]["month"]),
            int(data_item["internal"]["year"])
        )

        # Filtra datos por campo fecha
        date_column_filter = data_item['columns_date'][0]
        df_original = __get_data_filter_dates(
            df_original,
            date_column_filter,
            start_day,
            end_day,
            int(data_item["internal"]["month"]),
            int(data_item["internal"]["year"])
        )

        # df_compared = __get_data_filter_dates( df_compared, date_column_filter, start_day, end_day, int(data_item[
        # "internal"]["month"]), int(data_item["internal"]["year"]) )

        # TODO: Acá debiese filtrarse la información con
        #  respecto a la fecha definida dentro del ciclo generado -->
        #  df.loc[df['column_name'] == some_value]

        # Normaliza nombres de columnas / Eliminación de espacios en nombres
        item_error = f"Error al eliminar los espacios en las cabeceras fuente primaria"
        df_original = __remove_spaces_name_columns(df_original)
        item_error = f"Error al eliminar los espacios en las cabeceras fuente externa: {df_compared}"
        df_compared = __remove_spaces_name_columns(df_compared)

        # Normalización de los datos
        item_error = f"Error al normalizar los datos"
        for key, value in data_item['columns_normalized'].items():
            key = key.strip()
            df_original[f'{key}'] = __normalized_data(df_original, key, value)
            df_compared[f'{key}'] = __normalized_data(df_compared, key, value)

        # Definición de Índices
        item_error = f"Error al definir los indices"
        df_original = __create_index_in_df(df_original, 'idx', data_item['columns_index_df'])
        df_compared = __create_index_in_df(df_compared, 'idx', data_item['columns_index_df'])

        # Definición de campos a considerar dentro del DataFrame
        item_error = f"Error al definir los campos a usar en la consolidación"
        df_original = df_original[columns_df]
        df_compared = df_compared[columns_df]

        # Se ordenan los datos por fecha
        item_error = f"Error al ordenar la data por campo fecha"
        for key, value in data_item['columns_normalized'].items():
            key = key.strip()
            if value == 'datetime':
                df_original = __sort_datetime_data(df_original, key)
                df_compared = __sort_datetime_data(df_compared, key)

        # Se extraen los registros duplicados de cada origen
        item_error = f"Error al extraen los registros duplicados"
        df_duplicated_original = __get_duplicate(df_original)
        df_duplicated_compared = __get_duplicate(df_compared)

        # Elimina los registros duplicados
        item_error = f"Error al eliminar los registros duplicados"
        df_original = __delete_duplicate(df_original, data_item['columns_index_df'])
        df_compared = __delete_duplicate(df_compared, data_item['columns_index_df'])

        # Identificación de las transacciones conciliadas
        item_error = f"Error al identificar las transacciones conciliadas"
        df_reconciled = __merge_data(df_original, df_compared, 'idx')

        # Identificación de las transacciones no conciliadas
        item_error = f"Error al identificar las transacciones no conciliadas"
        df_unreconciled = __merge_data(df_reconciled, df_original, 'idx', False)

        # Identificación de las transacciones duplicadas
        item_error = f"Error al identificar las transacciones duplicadas"
        df_duplicated = __concat_data(df_duplicated_original, df_duplicated_compared, False)

        # Convierte campos datetime a str para conversion a JSON
        item_error = f"Error al tratar de convertir datos a texto"
        for key, value in data_item['columns_normalized'].items():
            key = key.strip()
            df_reconciled[f'{key}_primary'] = __convert_datetime_to_str(df_reconciled, f'{key}_primary', value)
            df_reconciled[f'{key}_external'] = __convert_datetime_to_str(df_reconciled, f'{key}_external', value)
            df_unreconciled[f'{key}_primary'] = __convert_datetime_to_str(df_unreconciled, f'{key}_primary', value)
            df_unreconciled[f'{key}_external'] = __convert_datetime_to_str(df_unreconciled, f'{key}_external', value)

            if key in df_reconciled:
                df_reconciled[f'{key}'] = __convert_datetime_to_str(df_reconciled, f'{key}', value)

            if key in df_unreconciled:
                df_unreconciled[f'{key}'] = __convert_datetime_to_str(df_unreconciled, f'{key}', value)

            if key in df_duplicated:
                df_duplicated[f'{key}'] = __convert_datetime_to_str(df_duplicated, f'{key}', value)

        # Normalización de campos vacíos (nulos)
        item_error = f"Error al tratar de normalizar campos vacíos (nulos)"
        for key, value in data_item['columns_normalized'].items():
            key = key.strip()
            df_reconciled[f'{key}_primary'] = __normalized_data(df_reconciled, f'{key}_primary', value)
            df_reconciled[f'{key}_external'] = __normalized_data(df_reconciled, f'{key}_external', value)
            df_unreconciled[f'{key}_primary'] = __normalized_data(df_unreconciled, f'{key}_primary', value)
            df_unreconciled[f'{key}_external'] = __normalized_data(df_unreconciled, f'{key}_external', value)
            df_duplicated[f'{key}'] = __normalized_data(df_duplicated, f'{key}', value)

        df_reconciled = __change_data_nat_nan_to_str(df_reconciled)
        df_unreconciled = __change_data_nat_nan_to_str(df_unreconciled)
        df_duplicated = __change_data_nat_nan_to_str(df_duplicated)

        # Agrega columna con descripción para identificación de problema en no conciliados.
        # __add_column_description(df_reconciled, df_unreconciled, data_item['columns_index_df'])

        # Generación de totales
        total_reconciled = 0
        total_unreconciled = 0
        total_duplicated = 0
        if len(data_item['columns_amount']) > 0:
            amount_column = data_item['columns_amount'][0]

            df_reconciled[f"{amount_column}_external"] = pd.to_numeric(df_reconciled[f"{amount_column}_external"])
            df_unreconciled[f"{amount_column}"] = pd.to_numeric(df_unreconciled[f"{amount_column}"])
            df_duplicated[f"{amount_column}"] = pd.to_numeric(df_duplicated[f"{amount_column}"])

            total_reconciled = df_reconciled[f"{amount_column}_external"].sum()
            total_unreconciled = df_unreconciled[f"{amount_column}"].sum()
            total_duplicated = df_duplicated[f"{amount_column}"].sum()

        # Se genera diccionario y se agrega campo de referencia para búsqueda
        item_error = f"Error al tratar de generar salida datos (Conciliados)"
        df_reconciled = df_reconciled.to_dict('records')
        idx = 1
        for df_rec in df_reconciled:
            df_rec[f'{settings.MONGO_COLLECTION_PROC_RECONCILED_CNF}_id'] = doc_id
            df_rec['item'] = idx
            df_rec['type_process'] = 'primary'
            idx += 1

        item_error = f"Error al tratar de generar salida datos (No Conciliados)"
        df_unreconciled = df_unreconciled.to_dict('records')
        idx = 1
        for df_unrec in df_unreconciled:
            df_unrec[f'{settings.MONGO_COLLECTION_PROC_RECONCILED_CNF}_id'] = doc_id
            df_unrec['item'] = idx
            df_unrec['type_process'] = 'primary'
            idx += 1

        df_duplicated = df_duplicated.to_dict('records')
        item_error = f"Error al tratar de generar salida datos (Duplicados)"
        idx = 1
        for df_dup in df_duplicated:
            df_dup[f'{settings.MONGO_COLLECTION_PROC_RECONCILED_CNF}_id'] = doc_id
            df_dup['item'] = idx
            df_dup['type_process'] = 'primary'
            idx += 1

        df_duplicated_compared = df_duplicated_compared.to_dict('records')
        item_error = f"Error al tratar de generar salida datos (Duplicados.)"
        idx = 1
        for df_dup_com in df_duplicated_compared:
            df_dup_com[f'{settings.MONGO_COLLECTION_PROC_RECONCILED_CNF}_id'] = doc_id
            df_dup_com['item'] = idx
            df_dup_com['type_process'] = 'primary'
            idx += 1

        result_process["primary-process"] = {
            "total_reconciled": total_reconciled,
            "data_reconciled": df_reconciled,
            "total_unreconciled": total_unreconciled,
            "data_unreconciled": df_unreconciled,
            "data_duplicated": df_duplicated,
            "total_duplicated": total_duplicated
        }

        # Proceso External v/s Primary
        file_name = f"{data_item['external'][0]['filename']}{data_item['external'][0]['extension']}"
        item_error = f"Error lectura archivo: {file_name}"
        df_original = __get_dataframe(
            constantes.H5_FILE,  # data_item['external'][0]['extension'],
            data_item['external'][0]['filename'],
            __getfile_from_path(sub_path, data_item['external'][0]['filename'], constantes.H5_FILE)
        )

        file_name = f"{data_item['internal']['filename']}{data_item['internal']['extension']}"
        item_error = f"Error lectura archivo: {file_name}"
        df_compared = __get_dataframe(
            constantes.H5_FILE,  # data_item['internal']['extension'],
            data_item['internal']['filename'],
            __getfile_from_path(sub_path, data_item['internal']['filename'], constantes.H5_FILE)
        )

        # Normaliza nombres de columnas / Eliminación de espacios en nombres
        item_error = f"Error al eliminar los espacios en las cabeceras fuente primaria"
        df_original = __remove_spaces_name_columns(df_original)
        item_error = f"Error al eliminar los espacios en las cabeceras fuente externa"
        df_compared = __remove_spaces_name_columns(df_compared)

        # Normalización de los datos
        item_error = f"Error al normalizar los datos"
        for key, value in data_item['columns_normalized'].items():
            key = key.strip()
            df_original[f'{key}'] = __normalized_data(df_original, key, value)
            df_compared[f'{key}'] = __normalized_data(df_compared, key, value)

        # Definición de Índices
        item_error = f"Error al definir los indices"
        df_original = __create_index_in_df(df_original, 'idx', data_item['columns_index_df'])
        df_compared = __create_index_in_df(df_compared, 'idx', data_item['columns_index_df'])

        # Definición de campos a considerar dentro del DataFrame
        item_error = f"Error al definir los campos a usar en la consolidación"
        df_original = df_original[columns_df]
        df_compared = df_compared[columns_df]

        # Se ordenan los datos por fecha
        item_error = f"Error al ordenar la data por campo fecha"
        for key, value in data_item['columns_normalized'].items():
            key = key.strip()
            if value == 'datetime':
                df_original = __sort_datetime_data(df_original, key)
                df_compared = __sort_datetime_data(df_compared, key)

        # Se extraen los registros duplicados de cada origen
        item_error = f"Error al extraen los registros duplicados"
        df_duplicated_original = __get_duplicate(df_original)
        df_duplicated_compared = __get_duplicate(df_compared)

        # Elimina los registros duplicados
        item_error = f"Error al eliminar los registros duplicados"
        df_original = __delete_duplicate(df_original, data_item['columns_index_df'])
        df_compared = __delete_duplicate(df_compared, data_item['columns_index_df'])

        # Identificación de las transacciones conciliadas
        item_error = f"Error al identificar las transacciones conciliadas"
        df_reconciled = __merge_data(df_original, df_compared, 'idx')

        # Identificación de las transacciones no conciliadas
        item_error = f"Error al identificar las transacciones no conciliadas"
        df_unreconciled = __merge_data(df_reconciled, df_original, 'idx', False)

        # Identificación de las transacciones duplicadas
        item_error = f"Error al identificar las transacciones duplicadas"
        df_duplicated = __concat_data(df_duplicated_original, df_duplicated_compared, False)

        # Convierte campos datetime a str para conversion a JSON
        item_error = f"Error al tratar de convertir datos a texto"
        for key, value in data_item['columns_normalized'].items():
            key = key.strip()
            df_reconciled[f'{key}_primary'] = __convert_datetime_to_str(df_reconciled, f'{key}_primary', value)
            df_reconciled[f'{key}_external'] = __convert_datetime_to_str(df_reconciled, f'{key}_external', value)
            df_unreconciled[f'{key}_primary'] = __convert_datetime_to_str(df_unreconciled, f'{key}_primary', value)
            df_unreconciled[f'{key}_external'] = __convert_datetime_to_str(df_unreconciled, f'{key}_external', value)

            if key in df_reconciled:
                df_reconciled[f'{key}'] = __convert_datetime_to_str(df_reconciled, f'{key}', value)

            if key in df_unreconciled:
                df_unreconciled[f'{key}'] = __convert_datetime_to_str(df_unreconciled, f'{key}', value)

            if key in df_duplicated:
                df_duplicated[f'{key}'] = __convert_datetime_to_str(df_duplicated, f'{key}', value)

        # Normalización de campos vacíos (nulos)
        item_error = f"Error al tratar de normalizar campos vacíos (nulos)"
        for key, value in data_item['columns_normalized'].items():
            key = key.strip()
            df_reconciled[f'{key}_primary'] = __normalized_data(df_reconciled, f'{key}_primary', value)
            df_reconciled[f'{key}_external'] = __normalized_data(df_reconciled, f'{key}_external', value)
            df_unreconciled[f'{key}_primary'] = __normalized_data(df_unreconciled, f'{key}_primary', value)
            df_unreconciled[f'{key}_external'] = __normalized_data(df_unreconciled, f'{key}_external', value)
            df_duplicated[f'{key}'] = __normalized_data(df_duplicated, f'{key}', value)

        df_reconciled = __change_data_nat_nan_to_str(df_reconciled)
        df_unreconciled = __change_data_nat_nan_to_str(df_unreconciled)
        df_duplicated = __change_data_nat_nan_to_str(df_duplicated)

        # Generación de totales
        total_reconciled = 0
        total_unreconciled = 0
        total_duplicated = 0
        if len(data_item['columns_amount']) > 0:
            amount_column = data_item['columns_amount'][0]

            df_reconciled[f"{amount_column}_external"] = pd.to_numeric(df_reconciled[f"{amount_column}_external"])
            df_unreconciled[f"{amount_column}"] = pd.to_numeric(df_unreconciled[f"{amount_column}"])
            df_duplicated[f"{amount_column}"] = pd.to_numeric(df_duplicated[f"{amount_column}"])

            total_reconciled = df_reconciled[f"{amount_column}_external"].sum()
            total_unreconciled = df_unreconciled[f"{amount_column}"].sum()
            total_duplicated = df_duplicated[f"{amount_column}"].sum()

        # Se genera diccionario y se agrega campo de referencia para búsqueda
        item_error = f"Error al tratar de generar salida datos (Conciliados)"
        df_reconciled = df_reconciled.to_dict('records')
        idx = 1
        for df_rec in df_reconciled:
            df_rec[f'{settings.MONGO_COLLECTION_PROC_RECONCILED_CNF}_id'] = doc_id
            df_rec['item'] = idx
            df_rec['type_process'] = 'secondary'
            idx += 1

        item_error = f"Error al tratar de generar salida datos (No Conciliados)"
        df_unreconciled = df_unreconciled.to_dict('records')
        idx = 1
        for df_unrec in df_unreconciled:
            df_unrec[f'{settings.MONGO_COLLECTION_PROC_RECONCILED_CNF}_id'] = doc_id
            df_unrec['item'] = idx
            df_unrec['type_process'] = 'secondary'
            idx += 1

        df_duplicated = df_duplicated.to_dict('records')
        item_error = f"Error al tratar de generar salida datos (Duplicados)"
        idx = 1
        for df_dup in df_duplicated:
            df_dup[f'{settings.MONGO_COLLECTION_PROC_RECONCILED_CNF}_id'] = doc_id
            df_dup['item'] = idx
            df_dup['type_process'] = 'secondary'
            idx += 1

        df_duplicated_compared = df_duplicated_compared.to_dict('records')
        item_error = f"Error al tratar de generar salida datos (Duplicados.)"
        idx = 1
        for df_dup_com in df_duplicated_compared:
            df_dup_com[f'{settings.MONGO_COLLECTION_PROC_RECONCILED_CNF}_id'] = doc_id
            df_dup_com['item'] = idx
            df_dup_com['type_process'] = 'secondary'
            idx += 1

        result_process["secondary-process"] = {
            "total_reconciled": total_reconciled,
            "data_reconciled": df_reconciled,
            "total_unreconciled": total_unreconciled,
            "data_unreconciled": df_unreconciled,
            "data_duplicated": df_duplicated,
            "total_duplicated": total_duplicated
        }

        return result_process

    except TypeError as te:
        log.error(te)
        return {
            "msg": f"{item_error}.\nError en el tipo de datos: {te}"
        }
    except KeyError as te:
        log.error(te)
        return {
            "msg": f"{item_error}.\nClave no existe: {te}."
        }
    except AttributeError as te:
        log.error(te)
        return {
            "msg": f"{item_error}.\nError en el atributo: {te}"
        }
    except InvalidIndexError as te:
        log.error(te)
        return {
            "msg": f"{item_error}.\nError en la definición del indice: {te}"
        }
    except ValueError as te:
        log.error(te)
        return {
            "msg": f"{item_error}.\nValor erróneo: {te}"
        }
    except Exception as e:
        log.exception(e)
        return {
            "msg": f"{item_error}.\nError: {e}"
        }


def convert_file_to_hfs5(path_file, file_name, extension_file):
    start = time.time()
    path_file_hfs5 = __getfile_from_path(path_file, file_name, constantes.H5_FILE)
    path_file_origin = __getfile_from_path(path_file, file_name, extension_file)

    # Carga de datos
    if extension_file == f'.{settings.FILE_EXT_CSV}':
        df = __get_dataframe(extension_file, file_name, path_file_origin)
    else:
        dfs = __read_excel(path_file_origin)
        df = pd.concat(dfs, ignore_index=True)

    df = df.astype(str)
    df = __remove_spaces_name_columns(df)

    # Convertir el DataFrame de Pandas a un DataFrame de Vaex
    df_vaex = vaex.from_pandas(df, copy_index=False)

    # Convertir todos los datos en cadenas de texto
    for columna in df_vaex.get_column_names():
        df_vaex[columna] = df_vaex[columna].astype('str')

    # Guardar el DataFrame de Vaex en formato HDF5
    df_vaex.export_hdf5(path_file_hfs5, mode='w', group=f'{file_name}')

    end = time.time()
    print(round((end - start) / 60, 2))
    try:
        os.remove(path_file_origin)
    except OSError:
        pass



import logging
import os

from bson import ObjectId
from flask import jsonify
from werkzeug.utils import secure_filename

from ..config import constantes, settings
from ..database.mongo_collections import (MONGO_COLLECTION_CONFIG_RULES,
                                          MONGO_COLLECTION_UPLOAD_FILE_PARTNER_ID,
                                          MONGO_COLLECTION_UPLOAD_FILE,
                                          MONGO_COLLECTION_CONFIG_RULES_DEFINED_COLUMNS,
                                          MONGO_COLLECTION_CONFIG_RULES_DEFINED_MAPPING_COLUMNS,
                                          MONGO_COLLECTION_CONFIG_RULES_DEFINED_INDEX_COLUMNS,
                                          MONGO_COLLECTION_CONFIG_RULES_DEFINED_AMOUNT_COLUMN,
                                          MONGO_COLLECTION_CONFIG_RULES_DEFINED_DATE_COLUMN,
                                          MONGO_COLLECTION_PROC_RECONCILED_FILE, MONGO_COLLECTION_PROC_RECONCILED_CNF,
                                          MONGO_COLLECTION_DATA_PARTNERS)
from ..database.mongodb import MongoAPI
from ..utility.utils import allowed_file, get_datetime, get_uuid, get_message_error
from ..utility.utils_file import extract_headers_file

log = logging.getLogger(__name__)


def create_config_rule(json: dict):
    """
    Método que permite crear una nueva Configuración de Regla
    :param json:
    :return:
    """
    try:
        data = {
            'collection': MONGO_COLLECTION_CONFIG_RULES,
            'Document': {
                f'{MONGO_COLLECTION_UPLOAD_FILE_PARTNER_ID}': json.get(
                    f'{MONGO_COLLECTION_UPLOAD_FILE_PARTNER_ID}'),
                'user_id': json.get('user_id'),
                'config_state': json.get('config_state'),
                'created_at': get_datetime(),
                'active': True
            }
        }
        mongodb = MongoAPI(data)
        response = mongodb.write(data, True)
        if response:
            response = {
                f'{MONGO_COLLECTION_UPLOAD_FILE_PARTNER_ID}': json.get(
                    f'{MONGO_COLLECTION_UPLOAD_FILE_PARTNER_ID}'),
                'user_id': json.get('user_id'),
                'config_state': json.get('config_state'),
                'config_rules_id': response['Document_ID']
            }
    except Exception as e:
        log.exception(e)
        response = get_message_error(e)
    return response


def update_config_rule(json: dict):
    try:
        data = {
            'collection': MONGO_COLLECTION_CONFIG_RULES,
            'Filter': {
                '_id': ObjectId(json.get('id')),
            },
            'DataToBeUpdated': {
                'config_state': json.get('config_state'),
                'name_rule': json.get('name_rule'),
                'updated_at': get_datetime()
            }
        }
        mongodb = MongoAPI(data)
        response = mongodb.update()
    except Exception as e:
        log.exception(e)
        response = get_message_error(e)
    return response


def validate_file(request):
    if 'file' not in request.files:
        resp = jsonify({'message': constantes.RESPONSE_NOT_PART_FILE_UPLOAD})
        resp.status_code = 400
        return resp

    file = request.files['file']
    '''if file.st_size == 0:
        resp = jsonify({'message': constantes.RESPONSE_EMPTY_SELECT_FILE_UPLOAD})
        resp.status_code = 400
    el'''
    if file.filename == constantes.EMPTY:
        resp = jsonify({'message': constantes.RESPONSE_NOT_SELECT_FILE_UPLOAD})
        resp.status_code = 400
    elif request.form['type'] is None:
        resp = jsonify({'message': constantes.RESPONSE_NOT_SETTING_TYPE_FILE_UPLOAD})
        resp.status_code = 400
    elif request.form['type'] == constantes.EMPTY:
        resp = jsonify({'message': constantes.RESPONSE_EMPTY_SETTING_TYPE_FILE_UPLOAD})
        resp.status_code = 400
    elif file and allowed_file(file.filename):
        file_name, file_extension = os.path.splitext(secure_filename(file.filename))

        # Save data upload
        type_data = request.form['type']
        user_id = request.form['user_id']
        partner_id = request.form[f'{MONGO_COLLECTION_UPLOAD_FILE_PARTNER_ID}']
        config_rule = request.form['config_rule']

        response = __set_upload_file(file, type_data, file_extension, user_id, partner_id, config_rule)
        if response:
            resp = jsonify(response)
            resp.status_code = 201
        else:
            resp = jsonify({'message': constantes.RESPONSE_ERROR_UPLOAD})
            resp.status_code = 500
    else:
        resp = jsonify({'message': f'Las extensiones permitidas son: {settings.FLASK_ALLOWED_EXTENSIONS}'})
        resp.status_code = 400

    return resp


def get_files(partner_id: str, config_rule_id: str):
    response = []
    try:
        data = {
            'collection': MONGO_COLLECTION_CONFIG_RULES,
            'Filter': {
                f'{MONGO_COLLECTION_UPLOAD_FILE_PARTNER_ID}': partner_id
            }
        }

        if config_rule_id != 'all':
            data['Filter']['_id'] = ObjectId(config_rule_id)

        mongodb = MongoAPI(data)
        config_rules = mongodb.all()
        for config_rule in config_rules:
            response = __get_response_fetch_upload(config_rule['_id'])

    except Exception as e:
        log.exception(e)
        response = get_message_error(e)

    return response


def get_file(pid: str):
    try:
        data = {
            'collection': MONGO_COLLECTION_UPLOAD_FILE,
            'Filter': {
                '_id': ObjectId(pid)
            }
        }
        mongodb = MongoAPI(data)
        response = mongodb.read()
    except Exception as e:
        log.exception(e)
        response = get_message_error(e)
    return response


def get_columns_file(pid: str, type_data: str = ''):
    response = {}
    try:
        data = {
            'collection': MONGO_COLLECTION_CONFIG_RULES_DEFINED_COLUMNS,
            'Filter': {
                f'{MONGO_COLLECTION_CONFIG_RULES}_id': pid
            }
        }

        if type_data != '':
            data['Filter']['type_data'] = type_data

        mongodb = MongoAPI(data)
        columns_file = mongodb.read()
        if len(columns_file) > 0:
            if type_data == '':
                data = {
                    'collection': MONGO_COLLECTION_UPLOAD_FILE,
                    'Filter': {
                        '_id': ObjectId(pid)
                    }
                }
                mongodb = MongoAPI(data)
                detail_file = mongodb.read()
                if len(detail_file) > 0:
                    response = {
                        'type': detail_file[0]['type'],
                        'columns': columns_file[0]['columns']
                    }
            else:
                response = {
                    'type': type_data,
                    'columns': columns_file[0]['columns']
                }

    except Exception as e:
        log.exception(e)
        response = get_message_error(e)
    return response


def set_columns_file(pid: str, json: dict):
    try:
        data = {
            'collection': MONGO_COLLECTION_CONFIG_RULES_DEFINED_COLUMNS,
            'Filter': {
                f'{MONGO_COLLECTION_CONFIG_RULES}_id': pid,
                'type_data': json.get('type')
            },
            'DataToBeUpdated': {
                'columns': json.get('columns')
            }
        }
        mongodb = MongoAPI(data)
        response = mongodb.update()
        if response:
            mapping = []
            for item in json.get('columns'):
                if bool(item.get('col_used')):
                    mapping.append({
                        'col_name': item.get('col_name'),
                        'col_name_mapped': ''
                    })

            if len(mapping) > 0:
                data = {
                    'collection': MONGO_COLLECTION_CONFIG_RULES_DEFINED_MAPPING_COLUMNS,
                    'Document': {
                        f'{MONGO_COLLECTION_CONFIG_RULES}_id': pid,
                        'type_data': json.get('type'),
                        'mapping': mapping
                    },
                    'Filter': {
                        f'{MONGO_COLLECTION_CONFIG_RULES}_id': pid,
                        'type_data': json.get('type')
                    },
                    'DataToBeUpdated': {
                        'mapping': mapping
                    }

                }
                mongodb = MongoAPI(data)
                response = mongodb.write(data)

            if response:
                response = __get_response_fetch_upload(pid)

    except Exception as e:
        log.exception(e)
        response = get_message_error(e)
    return response


def create_index_file(pid: str, json: dict):
    try:
        data = {
            'collection': MONGO_COLLECTION_CONFIG_RULES_DEFINED_INDEX_COLUMNS,
            'Document': {
                f'{MONGO_COLLECTION_CONFIG_RULES}_id': pid,
                'type_data': json.get('type'),
                'columns': json.get('columns')
            },
            'Filter': {
                f'{MONGO_COLLECTION_CONFIG_RULES}_id': pid,
                'type_data': json.get('type')
            },
            'DataToBeUpdated': {
                'columns': json.get('columns')
            }
        }
        mongodb = MongoAPI(data)
        response = mongodb.write(data)
        if response:
            response = __get_response_fetch_upload(pid)
    except Exception as e:
        log.exception(e)
        response = get_message_error(e)
    return response


def create_amount_column(pid: str, json: dict):
    try:
        data = {
            'collection': MONGO_COLLECTION_CONFIG_RULES_DEFINED_AMOUNT_COLUMN,
            'Document': {
                f'{MONGO_COLLECTION_CONFIG_RULES}_id': pid,
                'column': json.get('column')
            },
            'Filter': {
                f'{MONGO_COLLECTION_CONFIG_RULES}_id': pid
            },
            'DataToBeUpdated': {
                'column': json.get('column')
            }
        }
        mongodb = MongoAPI(data)
        response = mongodb.write(data)
        if response:
            response = __get_response_fetch_upload(pid)
    except Exception as e:
        log.exception(e)
        response = get_message_error(e)
    return response


def create_date_column(pid: str, json: dict):
    try:
        data = {
            'collection': MONGO_COLLECTION_CONFIG_RULES_DEFINED_DATE_COLUMN,
            'Document': {
                f'{MONGO_COLLECTION_CONFIG_RULES}_id': pid,
                'column': json.get('column')
            },
            'Filter': {
                f'{MONGO_COLLECTION_CONFIG_RULES}_id': pid
            },
            'DataToBeUpdated': {
                'column': json.get('column')
            }
        }
        mongodb = MongoAPI(data)
        response = mongodb.write(data)
        if response:
            response = __get_response_fetch_upload(pid)
    except Exception as e:
        log.exception(e)
        response = get_message_error(e)
    return response


def set_select_files_to_process(data: dict):
    pid = get_uuid()
    data = {
        'collection': MONGO_COLLECTION_PROC_RECONCILED_FILE,
        'Document': {
            f'{MONGO_COLLECTION_PROC_RECONCILED_CNF}_id': pid,
            'files': data.get('files')
        },
        'Filter': {
            f'{MONGO_COLLECTION_PROC_RECONCILED_CNF}_id': pid
        }
    }
    mongodb = MongoAPI(data)
    mongodb.write(data)


def get_index_file(pid: str, type_data: str):
    try:
        data = {
            'collection': MONGO_COLLECTION_CONFIG_RULES_DEFINED_INDEX_COLUMNS,
            'Filter': {
                f'{MONGO_COLLECTION_PROC_RECONCILED_CNF}_id': pid,
                'type_data': type_data
            }
        }
        mongodb = MongoAPI(data)
        response = mongodb.read()
    except Exception as e:
        log.exception(e)
        response = get_message_error(e)
    return response


def get_all_process_pending(cid: str):
    """
    Proceso que verifica si existen configuraciones pendientes
    :param cid: Identificador de la Configuración de la Regla
    :return: Lista de archivos pendientes por procesar y asociados al Socio
    """
    try:
        data = {
            'collection': MONGO_COLLECTION_UPLOAD_FILE,
            'Filter': {
                f'{MONGO_COLLECTION_CONFIG_RULES}_id': cid,
                'file_processed': False
            }
        }
        mongodb = MongoAPI(data)
        response = mongodb.read()
        if response:
            response = __get_response_fetch_upload(response)
    except Exception as e:
        log.error(e)
        response = get_message_error(e)
    return response


def get_all():
    response = []
    try:
        data = {
            'collection': MONGO_COLLECTION_CONFIG_RULES,
            'Filter': {
                'active': True
            }
        }
        mongodb = MongoAPI(data)
        config_rules = mongodb.all()
        for item in config_rules:
            data = {
                'collection': MONGO_COLLECTION_DATA_PARTNERS,
                'Filter': {
                    '_id': ObjectId(item[f'{MONGO_COLLECTION_UPLOAD_FILE_PARTNER_ID}'])
                }
            }
            mongodb = MongoAPI(data)
            partner = mongodb.read()
            if partner:
                if 'name_rule' not in item:
                    name_rule = f"Regla para Socio {partner[0]['name']}"
                    data = {
                        'collection': MONGO_COLLECTION_CONFIG_RULES,
                        'Filter': {
                            '_id': ObjectId(item['_id'])
                        },
                        'DataToBeUpdated': {
                            'name_rule': name_rule
                        }
                    }
                    mongodb = MongoAPI(data)
                    mongodb.update()
                else:
                    name_rule = item['name_rule']

                response.append({
                    f'{MONGO_COLLECTION_CONFIG_RULES}_id': item['_id'],
                    f'{MONGO_COLLECTION_UPLOAD_FILE_PARTNER_ID}': partner[0]['_id'],
                    'partner_name': partner[0]['name'],
                    'name_rule': name_rule,
                    'config_state': item['config_state'],
                    'created_at': item['created_at']
                })

    except Exception as e:
        log.error(e)
        response = get_message_error(e)
    return response


def set_mapping_columns(pid: str, json: dict):
    try:
        data = {
            'collection': MONGO_COLLECTION_CONFIG_RULES_DEFINED_MAPPING_COLUMNS,
            'Document': {
                f'{MONGO_COLLECTION_CONFIG_RULES}_id': pid,
                'type_data': json.get('type'),
                'mapping': json.get('mapping')
            },
            'Filter': {
                f'{MONGO_COLLECTION_CONFIG_RULES}_id': pid,
                'type_data': json.get('type')
            },
            'DataToBeUpdated': {
                'mapping': json.get('mapping')
            }
        }
        mongodb = MongoAPI(data)
        response = mongodb.write(data)
        if response:
            response = __get_response_fetch_upload(pid)
    except Exception as e:
        log.error(e)
        response = get_message_error(e)
    return response


def get_mapping_columns(pid: str):
    try:
        data = {
            'collection': MONGO_COLLECTION_CONFIG_RULES_DEFINED_MAPPING_COLUMNS,
            'Filter': {
                '_id': ObjectId(pid)
            }
        }
        mongodb = MongoAPI(data)
        response = mongodb.write(data)
    except Exception as e:
        log.error(e)
        response = get_message_error(e)
    return response


def get_all_config_rules_by_partner(cid: str, year: int = 0, month: int = 0, pid: str = ''):
    response = []
    try:
        data = {
            'collection': MONGO_COLLECTION_CONFIG_RULES,
            'Filter': {
                f'{MONGO_COLLECTION_UPLOAD_FILE_PARTNER_ID}': cid,
                'config_state': 'terminate',
                'active': True
            }
        }
        mongodb = MongoAPI(data)
        rules = mongodb.read()
        if rules:
            data = {
                'collection': MONGO_COLLECTION_PROC_RECONCILED_CNF,
                'Filter': {
                    f'{MONGO_COLLECTION_UPLOAD_FILE_PARTNER_ID}': cid,
                    # 'state': 'config'
                }
            }

            if len(pid) > 0:
                data['Filter']['_id'] = ObjectId(pid)

            if year > 0:
                data['Filter']['year'] = int(year)
                data['Filter']['month'] = int(month)

            mongodb = MongoAPI(data)
            config_proc = mongodb.read()

            if len(config_proc) > 0:
                rule = list(
                    filter(lambda r: r['_id'] != config_proc[0][f'{MONGO_COLLECTION_CONFIG_RULES}_id'], rules)
                )
                if rule:
                    response = rule
            else:
                response = rules
    except Exception as e:
        log.exception(e)
        response = get_message_error(e)
    return response


def config_remove(pid: str):
    try:
        data = {
            'collection': MONGO_COLLECTION_CONFIG_RULES,
            'Filter': {
                '_id': ObjectId(pid)
            }
        }
        mongodb = MongoAPI(data)
        response = mongodb.read()
        if len(response) > 0:
            data = {
                'collection': MONGO_COLLECTION_CONFIG_RULES,
                'Filter': {
                    '_id': ObjectId(pid)
                },
                'DataToBeUpdated': {
                    'active': False,
                    'updated_at': get_datetime()
                }
            }
            mongodb = MongoAPI(data)
            response = mongodb.delete(data)
    except Exception as e:
        log.exception(e)
        response = get_message_error(e)
    return response


##################################
#     Section Private Methods    #
##################################
def __set_upload_file(
        file,
        type_data: str,
        ext: str,
        user_id: str,
        partner_id: str,
        config_rule: str):
    try:
        time_process = get_datetime()
        response = __create_config_rule(config_rule, partner_id, user_id, time_process)
        if response:
            config_rules_id = response['Document_ID']
            columns = extract_headers_file(file, ext)
            if columns:
                response = __create_config_rule_source_columns(config_rules_id, type_data, columns)
                if len(response) > 0:
                    response = __get_response_fetch_upload(config_rules_id)

    except Exception as e:
        log.exception(e)
        response = get_message_error(e)
    return response


def __get_amount_column(pid: str):
    try:
        data = {
            'collection': MONGO_COLLECTION_CONFIG_RULES_DEFINED_AMOUNT_COLUMN,
            'Filter': {
                f'{MONGO_COLLECTION_CONFIG_RULES}_id': pid
            }
        }
        mongodb = MongoAPI(data)
        response = mongodb.read()
    except Exception as e:
        log.exception(e)
        response = get_message_error(e)
    return response


def __get_date_column(pid: str):
    try:
        data = {
            'collection': MONGO_COLLECTION_CONFIG_RULES_DEFINED_DATE_COLUMN,
            'Filter': {
                f'{MONGO_COLLECTION_CONFIG_RULES}_id': pid
            }
        }
        mongodb = MongoAPI(data)
        response = mongodb.read()
    except Exception as e:
        log.exception(e)
        response = get_message_error(e)
    return response


def __get_response_fetch_upload(cri: str, year: int = 0, month: int = 0):
    data = {
        'collection': MONGO_COLLECTION_CONFIG_RULES,
        'Filter': {
            '_id': ObjectId(cri)
        }
    }
    mongodb = MongoAPI(data)
    config_rules = mongodb.read()
    amount_column = []
    date_column = []
    for config_rule in config_rules:
        files = []
        partner_id = config_rule[f'{MONGO_COLLECTION_UPLOAD_FILE_PARTNER_ID}']
        data = {
            'collection': MONGO_COLLECTION_CONFIG_RULES_DEFINED_COLUMNS,
            'Filter': {
                f'{MONGO_COLLECTION_CONFIG_RULES}_id': cri
            }
        }

        if int(year) > 0 and int(month) > 0:
            data['Filter']['year'] = year
            data['Filter']['month'] = month

        mongodb = MongoAPI(data)
        response = mongodb.read()
        for item in response:
            columns = __get_response_fetch_upload_detail_columns(cri, item["type_data"])
            amount_column = __get_amount_column(cri)
            date_column = __get_date_column(cri)
            if item["type_data"] == 'external':

                data = {
                    'collection': MONGO_COLLECTION_CONFIG_RULES_DEFINED_MAPPING_COLUMNS,
                    'Filter': {
                        f'{MONGO_COLLECTION_CONFIG_RULES}_id': cri,
                        'type_data': item["type_data"]
                    }
                }
                mongodb = MongoAPI(data)
                response = mongodb.read()
                mapping = []

                if response:
                    mapping = response[0]['mapping']
                else:
                    for cols in columns:
                        if cols['col_used']:
                            mapping.append({
                                'col_name': cols['col_name'],
                                'col_name_mapped': ''
                            })

                files.append({
                    'type': f'{item["type_data"]}',
                    'columns': columns,
                    'mapping': mapping
                })
            else:
                data = {
                    'collection': MONGO_COLLECTION_CONFIG_RULES_DEFINED_INDEX_COLUMNS,
                    'Filter': {
                        f'{MONGO_COLLECTION_CONFIG_RULES}_id': cri,
                        'type_data': item["type_data"]
                    }
                }
                mongodb = MongoAPI(data)
                response = mongodb.read()
                indexes = []
                if response:
                    indexes = response[0]['columns']

                files.append({
                    'type': f'{item["type_data"]}',
                    'columns': columns,
                    'indexes': indexes
                })

        name_rule = ''
        if 'name_rule' in config_rule:
            name_rule = config_rule['name_rule']

        data_amount_col = ''
        data_date_col = ''
        if len(amount_column) > 0:
            data_amount_col = amount_column[0]['column']

        if len(date_column) > 0:
            data_date_col = date_column[0]['column']

        return {
            'id': cri,
            'name_rule': name_rule,
            f'{MONGO_COLLECTION_UPLOAD_FILE_PARTNER_ID}': partner_id,
            'config_rules': files,
            'amountColumn': data_amount_col,
            'amountDate': data_date_col
        }
    return {}


def __get_response_fetch_upload_detail_columns(config_rules_id: str, type_data: str):
    data = {
        'collection': MONGO_COLLECTION_CONFIG_RULES_DEFINED_COLUMNS,
        'Filter': {
            f'{MONGO_COLLECTION_CONFIG_RULES}_id': config_rules_id,
            'type_data': type_data
        }
    }
    mongodb = MongoAPI(data)
    upload_file_cols = mongodb.read()

    columns = []
    for cols in upload_file_cols[0]['columns']:
        columns.append({
            'col_name': cols['col_name'],
            'col_used': cols['col_used'],
            'col_norm': cols['col_norm'],
            'col_type': cols['col_type']
        })

    return columns


def __create_config_rule(config_rule: str, partner_id: str, user_id: str, time_process: str):
    # Crea Regla Socio
    data = {
        'collection': MONGO_COLLECTION_CONFIG_RULES,
        'Document': {
            'user_id': user_id,
            f'{MONGO_COLLECTION_UPLOAD_FILE_PARTNER_ID}': partner_id,
            'config_state': 'pending',
            'created_at': time_process
        },
        'Filter': {
            f'{MONGO_COLLECTION_UPLOAD_FILE_PARTNER_ID}': partner_id,
            '_id': ObjectId(config_rule)
        },
        'DataToBeUpdated': {
            'updated_at': time_process
        }
    }
    mongodb = MongoAPI(data)
    return mongodb.write(data)


def __create_upload_file_columns(config_rules_id: str, columns):
    # Carga Columnas de Archivo Subido
    data = {
        'collection': MONGO_COLLECTION_CONFIG_RULES_DEFINED_COLUMNS,
        'Document': {
            f'{MONGO_COLLECTION_CONFIG_RULES}_id': config_rules_id,
            'columns': columns
        },
        'Filter': {
            f'{MONGO_COLLECTION_CONFIG_RULES}_id': config_rules_id
        }
    }
    mongodb = MongoAPI(data)
    return mongodb.write(data)


def __create_config_rule_source_columns(config_rules_id: str, type_data: str, columns):
    """
    Carga Columnas de Archivo Subido
    :param config_rules_id:
    :param columns:
    :return:
    """
    data = {
        'collection': MONGO_COLLECTION_CONFIG_RULES_DEFINED_COLUMNS,
        'Document': {
            f'{MONGO_COLLECTION_CONFIG_RULES}_id': config_rules_id,
            'type_data': type_data,
            'columns': columns
        },
        'Filter': {
            f'{MONGO_COLLECTION_CONFIG_RULES}_id': config_rules_id,
            'type_data': type_data
        },
        'DataToBeUpdated': {
            'columns': columns
        }
    }
    mongodb = MongoAPI(data)
    return mongodb.write(data)



from ..config import settings
from ..database import MongoAPI
from ..database.mongo_collections import MONGO_COLLECTION_AUDITORY_LOGS
from ..enums.enums import EventsLogsEnum
from ..utility import get_datetime


class AuditoryLogs:
    @staticmethod
    def registry_log(origin: str, event: EventsLogsEnum, description: str, user: str):
        data = {
            'collection': MONGO_COLLECTION_AUDITORY_LOGS,
            'Document': {
                'origin': origin,
                'event': event.value,
                'description': description,
                'user_id': user,
                'created_at': get_datetime()
            },
            'Filter': {
                'origin': origin,
                'description': description,
            }
        }

        mongodb = MongoAPI(data)
        mongodb.write(data)



import json
import logging
import ssl
import sys
import traceback

from kafka import KafkaProducer, KafkaConsumer
from kafka.errors import KafkaError
from retrying import retry

from ..config import settings

log = logging.getLogger(__name__)


class KafkaConfig:
    producer = None
    consumer = None

    def __get_context(self):
        # Create a new context using system defaults, disable all but TLS1.2
        self.context = ssl.create_default_context()
        self.context.options &= ssl.OP_NO_TLSv1
        self.context.options &= ssl.OP_NO_TLSv1_1

        return self.context

    @retry(wait_exponential_multiplier=500, wait_exponential_max=10000)
    def __set_producer(self, is_ssl=False):
        try:
            brokers = f'{settings.KAFKA_BOOSTRAP}:{settings.KAFKA_PORT}'
            log.info(f"Creating new kafka producer using brokers: {str(brokers)}")
            if is_ssl:
                self.producer = KafkaProducer(
                    bootstrap_servers=brokers,
                    sasl_plain_username=settings.KAFKA_USERNAME,
                    sasl_plain_password=settings.KAFKA_PASSWORD,
                    security_protocol=settings.KAFKA_SECURITY_PROTOCOL,
                    ssl_context=self.__get_context(),
                    sasl_mechanism=settings.KAFKA_SASL_MECHANISM,
                    value_serializer=lambda m: json.dumps(m).encode('utf-8'),
                    api_version=(0, 10, 1),
                    retries=5
                )
            else:
                self.producer = KafkaProducer(
                    bootstrap_servers=brokers,
                    value_serializer=lambda m: json.dumps(m).encode('utf-8'),
                    api_version=(0, 10, 1),
                    retries=5,
                    max_block_ms=5000
                )
        except KeyError as e:
            log.exception('Missing setting named ' + str(e), {'ex': traceback.format_exc()})
        except Exception:
            log.exception("Couldn't initialize kafka producer.", {'ex': traceback.format_exc()})
            raise

    @retry(wait_exponential_multiplier=500, wait_exponential_max=10000)
    def __set_consumer(self, topic, group_id):
        try:
            brokers = f'{settings.KAFKA_BOOSTRAP}:{settings.KAFKA_PORT}'
            log.info(f'Creating new kafka consumer using brokers: {str(brokers)} and topic {topic}')
            self.consumer = KafkaConsumer(
                topic,
                group_id=group_id,
                bootstrap_servers=brokers,
                auto_offset_reset=settings.KAFKA_OFFSET,
                consumer_timeout_ms=settings.KAFKA_CONSUMER_TIMEOUT,
                enable_auto_commit=settings.KAFKA_CONSUMER_AUTO_COMMIT_ENABLE,
                value_deserializer=json.loads
            )
        except KeyError as e:
            log.exception('Missing setting named ' + str(e), {'ex': traceback.format_exc()})
        except Exception:
            log.exception("Couldn't initialize kafka consumer for topic", {'ex': traceback.format_exc(), 'topic': topic})
            raise

    def send_message(self, topic, message):
        try:
            self.__set_producer()
            if self.producer is not None:
                self.producer.send(topic, message)
                self.producer.flush()
                self.producer.close(timeout=10)
        except KafkaError as e:
            log.exception("Unexpected error:" + str(e), sys.exc_info()[0])
            raise

    def get_message(self, topic, group_id):
        try:
            self.__set_consumer(topic, group_id)
            return self.consumer
        except KafkaError as e:
            log.exception("Unexpected error:" + str(e), sys.exc_info()[0])
            raise



import logging
import platform
import traceback

import oracledb
from oracledb import DatabaseError
from sqlalchemy import create_engine, NullPool
from sqlalchemy.orm import sessionmaker

from ..config import settings

log = logging.getLogger(__name__)

d = None  # default suitable for Linux
f = None
if platform.system() == 'Darwin' and platform.machine() == 'x86_64':  # macOS
    d = 'linux'
    f = '.'
elif platform.system() == 'Windows':
    d = 'win'
    f = 'c:'
oracledb.init_oracle_client(lib_dir=f'{f}/oracle_client/{d}')

log = logging.getLogger(__name__)


class OracleAPI:
    def __init__(self, show_sql=False):
        try:
            if settings.ORACLE_SERVICE_NAME != '':
                cp = oracledb.ConnectParams(
                    host=f'{settings.ORACLE_HOST}',
                    port=int(settings.ORACLE_PORT),
                    sid=f'{settings.ORACLE_SERVICE_NAME}',
                    user=settings.ORACLE_USERNAME,
                    password=settings.ORACLE_PASSWORD
                )
            elif settings.ORACLE_DB_NAME != '':
                cp = oracledb.ConnectParams(
                    host=f'{settings.ORACLE_HOST}',
                    port=int(settings.ORACLE_PORT),
                    service_name=f'{settings.ORACLE_DB_NAME}',
                    user=settings.ORACLE_USERNAME,
                    password=settings.ORACLE_PASSWORD
                )

            dsn = cp.get_connect_string()

            self.pool = oracledb.create_pool(
                user=settings.ORACLE_USERNAME,
                password=settings.ORACLE_PASSWORD,
                dsn=dsn,
                min=1,
                max=4,
                increment=1,
                disable_oob=True,
                ping_interval=2,
                getmode=oracledb.POOL_GETMODE_WAIT
            )

            self.engine = create_engine(
                f'oracle+oracledb://:@',
                creator=self.pool.acquire,
                poolclass=NullPool,
                echo=show_sql,
                thick_mode=False,
                connect_args={
                    'user': settings.ORACLE_USERNAME,
                    'password': settings.ORACLE_PASSWORD,
                    'host': settings.ORACLE_HOST,
                    'port': settings.ORACLE_PORT,
                    'service_name': settings.ORACLE_DB_NAME,
                    'dsn': dsn,
                    'dbname': settings.ORACLE_SCHEMA,
                    'schema': settings.ORACLE_SCHEMA,
                    'options': f'-csearch_path={settings.ORACLE_SCHEMA}'
                }
            )

            # Create a session to the database
            Session = sessionmaker(bind=self.engine)
            self.session = Session()

        except DatabaseError:
            log.exception("Couldn't connect to Oracle DataBase.", {'ex': traceback.format_exc()})

    def read(self, sql):
        try:
            result = self.session.execute(sql).fetchone()
        except Exception as ex:
            log.exception(ex)
            result = []
        return result if len(result) > 0 else []

    def all(self, sql):
        try:
            result = self.session.execute(sql).fetchall()
        except Exception as ex:
            log.exception(ex)
            result = []
        return result if len(result) > 0 else []



# Version liberación - Release ddmmyyyy
from ..services.config_params import get_config_params

APP_VERSION = '1.19.0'
# Flask Settings
FLASK_SECRET_KEY = '0pt1m1$42560-$3cr3t4'

FLASK_DEBUG = True
FLASK_THREADED = True
FLASK_ERROR_404_HELP = False
FLASK_APP_FOLDER = 'optidata-core'
FLASK_NAME_UPLOAD_FOLDER = 'files'
FLASK_NAME_EXPORT_FOLDER = 'exported-files'
FLASK_ALLOWED_EXTENSIONS = {'txt', 'csv', 'xls', 'xlsx'}
FLASK_MAX_CONTENT_LENGTH = 500 * 1024 * 1024
FLASK_NUMBER_PROCESSES = 2
FLASK_ROOT_PATH = ''
FLASK_UPLOAD_FOLDER = ''
FLASK_EXPORT_FOLDER = ''

# Flask-Restplus settings
RESTPLUS_SWAGGER_UI_DOC_EXPANSION = 'list'
RESTPLUS_VALIDATE = True
RESTPLUS_MASK_SWAGGER = False
RESTPLUS_ERROR_404_HELP = False
RESTPLUS_API_VERSION = '/v1'
RESTPLUS_EMPTY_RESULT = 'No existen registros asociados'

# Oracle Properties (SAP)
ORACLE_HOST = get_config_params('ORACLE_HOST', 'database')
ORACLE_PORT = get_config_params('ORACLE_PORT', 'database')
ORACLE_USERNAME = get_config_params('ORACLE_USER', 'database')
ORACLE_PASSWORD = get_config_params('ORACLE_PWD', 'database')
ORACLE_DB_NAME = get_config_params('ORACLE_DB', 'database')
ORACLE_SERVICE_NAME = get_config_params('ORACLE_SID', 'database')
ORACLE_SCHEMA = get_config_params('ORACLE_SCHEMA', 'database')
ORACLE_ENCODING = 'UTF-8'

# BCrypt Config
BCRYPT_LOG_ROUNDS = 13

# JWT Configuration
TOKEN_EXPIRE_HOURS = 1
TOKEN_EXPIRE_MINUTES = 15
TOKEN_REFRESH_EXPIRE_HOURS = 1
TOKEN_HEADER_NAME = 'X-Access-Token'
TOKEN_FORMAT = 'Bearer '

# Files Configuration
FILE_EXT_TXT = 'txt'
FILE_EXT_CSV = 'csv'
FILE_EXT_XLS = 'xls'
FILE_EXT_XLSX = 'xlsx'
CSV_SEPARATOR = get_config_params('CSV_SEPARATOR', 'default')

# Vaex Configuration
VAEX_THREAD_COUNT = 10
VAEX_MAX_COLUMN = 10000

# Default Configurations - User/Role
DEFAULT_USERNAME = 'admin@optimisa.cl'
DEFAULT_PASSWORD = ''
DEFAULT_FIRSTNAME = 'Administrador'
DEFAULT_LASTNAME = 'del Sistema'

DEFAULT_USER_USERNAME = 'consulta@optimisa.cl'
DEFAULT_USER_PASSWORD = ''
DEFAULT_USER_FIRSTNAME = 'Usuario Consulta'
DEFAULT_USER_LASTNAME = 'del Sistema'

DEFAULT_ROLE_NAME_ADMIN = 'Administrador'
DEFAULT_ROLE_NAME_USER = 'Consulta'

# Kafka Properties
KAFKA_BOOSTRAP = get_config_params('KAFKA_SERVER', 'kafka')
KAFKA_PORT = get_config_params('KAFKA_PORT', 'kafka')
KAFKA_SASL_MECHANISM = 'PLAIN'
KAFKA_SECURITY_PROTOCOL = 'SASL_SSL'
KAFKA_USERNAME = get_config_params('KAFKA_USER', 'kafka')
KAFKA_PASSWORD = get_config_params('KAFKA_PWD', 'kafka')
KAFKA_OFFSET = 'earliest'
KAFKA_TIMEOUT = 1000
KAFKA_ENCODE = 'UTF-8'

KAFKA_CONSUMER_COMMIT_INTERVAL_MS = 5000
KAFKA_CONSUMER_TIMEOUT = 50
KAFKA_CONSUMER_FETCH_MESSAGE_MAX_BYTES = 10 * 1024 * 1024  # 10MB
KAFKA_CONSUMER_AUTO_COMMIT_ENABLE = True

# Kafka Topics
KAFKA_TOPIC_SCHEDULER = 'topic_schedulers'
KAFKA_TOPIC_PETITION = 'topic-requeriments'

# SFTP Properties
SFTP_HOSTNAME = get_config_params('SFTP_HOST', 'sftp')
SFTP_PORT = get_config_params('SFTP_PORT', 'sftp')
SFTP_USERNAME = get_config_params('SFTP_USER', 'sftp')
SFTP_PASSWORD = get_config_params('SFTP_PWD', 'sftp')
SFTP_PATH_DEFAULT = 'sftp'

# Scheduler Properties
SCHEDULER_MIN_TIME_RUN = 1
SCHEDULER_MIN_TIME_RUN_LOTES = 1
SCHEDULER_MIN_TIME_RUN_CARTOLA = 1
SCHEDULER_MIN_TIME_RUN_PA = 1



# MongoDB Configuration
MONGO_BATCH_SIZE = 100000
MONGO_DB_HOST = '10.10.1.128'  # get_config_params('DB_HOST')
MONGO_DB_PORT = '27027'  # get_config_params('DB_PORT')
MONGO_DB_NAME = 'optidata_db'
MONGO_DB_URI = f'mongodb://{MONGO_DB_HOST}:{MONGO_DB_PORT}/{MONGO_DB_NAME}'
MONGO_DB_USER = 'admin@optimisa.cl'  # get_config_params('DB_USER')
MONGO_DB_PWD = '0pt1m1542560_2024'  # get_config_params('DB_PWD')
